
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package TreeJuxtaposer;

import java.util.*;
import java.awt.event.*;

import AccordionDrawer.*;
import AccordionTreeDrawer.*;

/**
 * A class that fleshes out the abstract methods left in
 * AccordionTreeDrawer.
 *
 * In particular, mouse and keyboard events are mapped to
 * AccordionTreeDrawer methods.
 *
 * @author  Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.AccordionDrawer
 * @see     AccordionDrawer.AccordionTreeDrawer
 */

public class AccordionTreeDrawerFinal extends AccordionTreeDrawer {

    TreeJuxtaposer tj;
    int dragStart[] = new int[2];
    int dragEnd[] = new int[2];
    int mousePrev[]=new int[2];
    int mouseNow[]=new int[2];
    TreeNode mouseOverNode;

	// actiongroup: what is supposed to grow when grow option (bigger/smaller) is pressed
    private int actiongroup;
    // markgroup: what is the selection colour that is currently being used
    //  (only applies to MARK[0-3], other ints are only associated with actiongroup)
    private int markgroup;
    /*
     * Don't use these and expect it to work since tj.????Group stuff is used elsewhere
    final static int MARK0 = 0;
    final static int MARK1 = 1;
    final static int MARK2 = 2;
    final static int MARK3 = 3;
    final static int FLASH = 4;
    final static int DIFF = 5;
    final static int LIST = 6;
    */
    
    // which direction to grow
    private int growDirection;
    final static int HORIZ = 0;
    final static int VERT = 1;
    final static int ALL = 2;
    private int growDirectionDefault = VERT;
    
    // does the selection (marking) of a node select just the node or the subtree
    private int selectionResolution;
    final static int NODE = 0;
    final static int SUBTREE = 1;
    private int selectionResolutionDefault = SUBTREE;
    
	private int keytarget;
    // developer mode for keytarget only
    final static int DEVELOPER = 11;
    // only let the keytarget and actionmode be set to NONE
    // any other state that can be seen externally should have real state (default value)
    final static int NONE = 12;
    private final static int keytargetDefault = TreeJuxtaposer.flashGroup;
    
    private int actionmode;
    final static int MOUSEOVER = 0;
    final static int ST_FREEMOVE = 1;
    final static int ST_FREEMOVEAGAIN = 2;
    final static int ST_RESHAPE = 3;
    final static int RECT_CREATE = 4;
    final static int RECT_FREEMOVE = 5;
    final static int RECT_FREEMOVEAGAIN = 6;
    final static int RECT_RESHAPE = 7;
    private final static int actionmodeDefault = MOUSEOVER;
	final static int MOVE_STUCKPOS = 11;

    boolean jump;
    TreeNode pickedNode;
    TreeNode pickedNodeOld;

    

	/* State accessors and mutators */
	
	/* Action group */
	public void setActionGroup(int newActionGroup)
	{
		//System.out.println("setting action group: " + newActionGroup);
		setActionGroupWrapper(newActionGroup, true);
	}
	public void setActionGroupWrapper(int newActionGroup, boolean repeat)
	{
		actiongroup = newActionGroup;
		if (repeat)
		{
			Iterator tdIter = tj.treeDrawers.iterator();
			while (tdIter.hasNext())
			{
				AccordionTreeDrawerFinal atd = (AccordionTreeDrawerFinal)tdIter.next();
				atd.setActionGroupWrapper(newActionGroup, false);
			}
		}
		tj.observe();
	}
	public int getActionGroup()
	{
		return actiongroup;
	}
	
	/* Action mode */
	public void setActionMode(int newActionMode)
	{
		//System.out.println("setting action mode: " + newActionMode);
		setActionModeWrapper(newActionMode, false);
	}
	public void setActionModeWrapper(int newActionMode, boolean repeat)
	{
		actionmode = newActionMode;
		if (repeat)
		{
			Iterator tdIter = tj.treeDrawers.iterator();
			while (tdIter.hasNext())
			{
				AccordionTreeDrawerFinal atd = (AccordionTreeDrawerFinal)tdIter.next();
				atd.setActionModeWrapper(newActionMode, false);
			}
		}
		tj.observe();
	}
	public int getActionMode()
	{
		return actionmode;
	}
	
	public void setGrowDirection(int growDirection)
	{
		//System.out.println("setting grow direction: " + growDirection);
		setGrowDirectionWrapper(growDirection, true);
	}
	public void setGrowDirectionWrapper(int growDirection, boolean repeat)
	{
		this.growDirection = growDirection;
		if (repeat)
		{
			Iterator tdIter = tj.treeDrawers.iterator();
			while (tdIter.hasNext())
			{
				AccordionTreeDrawerFinal atd = (AccordionTreeDrawerFinal)tdIter.next();
				atd.setGrowDirectionWrapper(growDirection, false);
			}
		}
		tj.observe();
	}
	public int getGrowDirection()
	{
		return growDirection;
	}
	
	public void setSelectionResolution(int selectionResolution)
	{
		//System.out.println("setting action target: " + selectionResolution);
		setSelectionResolutionWrapper(selectionResolution, true);
	}
	public void setSelectionResolutionWrapper(int selectionResolution, boolean repeat)
	{
		this.selectionResolution = selectionResolution;
		if (repeat)
		{
			Iterator tdIter = tj.treeDrawers.iterator();
			while (tdIter.hasNext())
			{
				AccordionTreeDrawerFinal atd = (AccordionTreeDrawerFinal)tdIter.next();
				atd.setSelectionResolutionWrapper(selectionResolution, false);
			}
		}
		tj.observe();
	}
	public int getSelectionResolution()
	{
		return selectionResolution;
	}
	
	public void setMarkGroup(int newMarkGroup)
	{
		setMarkGroupWrapper(newMarkGroup, true);
	}
	public void setMarkGroupWrapper(int newMarkGroup, boolean repeat)
	{
		markgroup = newMarkGroup;
		if (repeat)
		{
			Iterator tdIter = tj.treeDrawers.iterator();
			while (tdIter.hasNext())
			{
				AccordionTreeDrawerFinal atd = (AccordionTreeDrawerFinal)tdIter.next();
				atd.setMarkGroupWrapper(newMarkGroup, false);
			}
		}
		tj.observe();
	}
	public int getMarkGroup()
	{
		return markgroup;
	}

    AccordionTreeDrawerFinal(Tree t, int w, int h, TreeJuxtaposer thetj) {
	super(t, w, h);
	tj = thetj;
	setMarkGroup(TreeJuxtaposer.markGroup[0]);
	setActionGroup(TreeJuxtaposer.flashGroup);
	setActionMode(actionmodeDefault);
	keytarget = NONE;
	setGrowDirection(growDirectionDefault);
	setSelectionResolution(selectionResolutionDefault);
	jump = false;
	baseBox = null;
	splitLine[X].setMaxStuckValue(0.7f); // so you can see the names on the leaves
    }

    // implement abstract classes
    public ArrayList getColorsForRange(int objmin, int objmax) {
	return tj.getColorsForRange(objmin, objmax, (AccordionTreeDrawer) this);
    }

	// implement abstract classes
	public ArrayList getColorsForRange(int objmin, int objmax, int a, int b) {
	   return tj.getColorsForRange(objmin, objmax, (AccordionTreeDrawer) this);
	   }

    /**
     * Make an InteractionBox from the GridCells nearest to the box of the mousedrag.
     * Expects dragStart[X]/Y, dragEnd[X]/Y to be set. 
     * 
     * @author Francois Guimbretiere, Tamara Munzner */

    private void drawActiveSubtreeBox() {
	// turn off original flashbox
	flashBox = null;
	setFlash(null, rubberbandColor, -1, -1, true);
	CellGeom fg = null;
	if (getActionGroup() == TreeJuxtaposer.flashGroup) {
	    baseBox = makeBox(pickedNode); 
	    if (null == baseBox) {
//		System.out.println("basebox null");
		return;
	    }
	    flashBox = baseBox;
	    if (null != pickedNode)
			fg = pickedNode;
//	    else if (null != temp && null != temp.item)
//		fg = temp.item;
	} else {
	    ArrayList forestRoots = tj.getGroupForest(getActionGroup(), this); // array of TreeNodes
	    if (forestRoots.isEmpty()) {
		actionmodeReset();
		return;
	    }
	    TreeNode n = (TreeNode)(forestRoots.get(0));
	    baseBox = makeBox(n); // just use the first one, for now
	    if (null == baseBox) {
//		System.out.println("basebox2 null");
		return;
	    }
	    fg = n;
	    flashBox = baseBox;
	}
	flashBoxWidth=3f;
	if (fg == null) fg = tree.getRoot();
	tj.setFocus(fg, this);
	setFlash(fg, rubberbandColor,-1, -1, true);
    }

    public void actionmodeReset() {
	//System.out.println("reset");
	
	// reason: set action mode to mouseover when exiting other modes
	setActionMode(MOUSEOVER);
	
	baseBox = null;
	flashBox = null;
	pickedNode = null;
	pickedNodeOld = null;
	flashBoxWidth = 1f;
	tj.doFlashGeom(null, TreeJuxtaposer.flashGroup, this, mouseNow[X], mouseNow[Y], false);
	//setFlash(null, tj.rubberbandColor, mouseNow[X], mouseNow[Y], true);    
	// can't set this null until after setflash...
	//flashBoxOld = null;
	//flashGeomOld = null;
	//flashGeom = null;
	
	// reason?
	// setActionGroup(tj.flashGroup);
	
	keytarget = NONE; // squash keys in state machine
	//setGrowDirection(growDirectionDefault); // this resets the grow direction to both... why?
	tj.setQuasimode(false);
	mouseover(mouseNow[X], mouseNow[Y]);
    }

    private void mouseover(int x, int y) {     
	TreeNode pickedGeom = (TreeNode)pickAttached(x,y);
	if (null == pickedGeom) return;
        mouseOverNode = pickedGeom;
    pickedNode = pickedGeom;
	if (pickedNode == pickedNodeOld) return;
	

	pickedNodeOld = pickedNode;
	tj.doFlashGeom(pickedNode, TreeJuxtaposer.flashGroup, this, x, y, true);
	tj.debugFrame.result[DebugFrame.NAV_TYPE].setText(pickedNode.toString());
	tj.debugFrame.result[DebugFrame.BCN_SCORE].setText((pickedNode.getBcnScore()).toString());
    
    }

	private void reshaperectangle(boolean doOther) {
		
		if (null == baseBox) {
			actionmodeReset();
			return;
		}
		
		// CRFP: return 2 integers, pixel differences for baseBox movable lines in X and Y
		int[] rectBox = createRectFromPick(baseBox);
		
		if (rectBox == null)
			return;
//		System.out.println("RectBox: " + rectBox[0] + " " + rectBox[1]);	
		int numAnimSteps = 3;
		while (transitionAnimating)
			endAllTransitions(); // end transitions, new queue is going to be swapped in
		Hashtable newToMove = new Hashtable();
		for (int xy = X; xy <= Y; xy++)
		{
			splitLine[xy].moveLine(baseBox.moveLineIndex[xy], // the split to move
				rectBox[xy], // move it to the end position
				baseBox.stuckLineIndex[xy], // the split to stay
				numAnimSteps, newToMove);
		}
		toMove = newToMove; // swap in new queue
		incrementFrameNumber(); // so compute place this frame will update, this can be moved to after transitions are created
		

		if (doOther
			&& null != baseBox.item) { // linked navigation to other trees

			double changeRatio[] = new double[2];
			for (int xy = 0; xy < 2; xy++) {
				SplitLine split = getSplitLine(xy);
				int movement = baseBox.dragEnd[xy] - baseBox.oldDragEnd[xy];
//				System.out.println("movement: " + movement);
				if (baseBox.stuckLineIndex[xy] > baseBox.moveLineIndex[xy])
					movement *= -1;
				changeRatio[xy] = s2w(movement, xy);
//				double dragEnd = (double)s2w(rectBox[xy], xy);
//				if (dragEnd < split.getMinStuckValue() + minContextPeriphery ||
//					dragEnd > split.getMaxStuckValue() - minContextPeriphery)
//				{
//					System.out.println("drag too far");
//					break;
//					
//				}
//				double staticEnd = split.getAbsoluteValue(baseBox.stuckLineIndex[xy]);
//				double dragStart = split.getAbsoluteValue(baseBox.moveLineIndex[xy]);
//				double oldBoxSize = Math.abs(staticEnd - dragStart);
//				double newBoxSize = Math.abs(staticEnd - dragEnd);
//
//
//				changeRatio[xy] = newBoxSize - oldBoxSize;
//
			}
			TreeNode n = (TreeNode)baseBox.item;
			tj.resizeRectOthers(n, changeRatio, numAnimSteps, this);
			tj.requestRedrawAll();
		}
		else
			requestRedraw(); // redraw the frame, this will enact the movements in the move queue
		flashBox = baseBox;
		setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
	}

    private void remakerectangle() {
//	if (null == flashBox) {
//	    actionmodeReset(); return;
//	}
//	// flashbox is leftover from last reshaperectangle.
//	/*
//	// need to transfer the cells from the last basebox to it.
//	flashBox.minc = baseBox.minc;
//	flashBox.maxc = baseBox.maxc;
//	System.out.println("remake "+flashBox.minc+" "+flashBox.maxc); 
//	setFlash(null, tj.rubberbandColor, -1, -1, true);
//	return;
//	*/
//	// need to remake the baseBox with real cell data
//	// using boundaries of last rectmove flashBox
//	for (int xy = 0; xy<2; xy++) {
//	    // get just inside box
//	    dragStart[xy] = flashBox.min[xy]+1;
//	    dragEnd[xy] = flashBox.max[xy]-1;
//	}
//	CellGeom item = baseBox.item;
//	baseBox = createBoxFromCells(dragStart, dragEnd);
//	baseBox.item = item;
//	if (baseBox.minc == null || baseBox.maxc == null) System.out.println("NO CELLS!");
//	flashBox = baseBox;
//	setFlash(null, rubberbandColor, -1, -1, true);
//	// put back dragStart/End to real values
//	for (int xy = 0; xy < 2; xy++) {
//	    dragStart[xy] = mouseNow[xy];
//	    dragEnd[xy] = mouseNow[xy];
//	}
    }

    public void mouseEntered(MouseEvent e) {
    	requestRedraw();
		if (mouseOutOfWindow(e))
			return;
	//System.out.println("entered window");
	if (!tj.getQuasimode()) requestFocus();
	else tj.wantsFocusInQuasi(this);
    }
    public void mouseExited(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
	//System.out.println("left window");
	if (!tj.getQuasimode()) transferFocus();
	//else tj.lostFocusInQuasi(this);
    }

	private boolean mouseOutOfWindow(MouseEvent e)
	{
		return e.getX() <= 0 || e.getX() >= getWinMax(X) || e.getY() <= 0 || e.getY() >= getWinMax(Y);
	}

	private int closeToStuck()
	{
		double mouse[] = {s2w(mouseNow[X], X), s2w(mouseNow[Y], Y)};
		double fuzz[] = {s2w(pickFuzz, Y), s2w(pickFuzz, Y)};
		for (int xy = X; xy <= Y; xy++)
		{
			if (Math.abs(mouse[xy] - splitLine[xy].getMinStuckValue()) < fuzz[xy])
			{
				//System.out.println("close to minstuck? " + xy);
				return xy + MIN_STUCK_X;
			}
			else if (Math.abs(mouse[xy] - splitLine[xy].getMaxStuckValue()) < fuzz[xy])
			{
				//System.out.println("close to maxstuck? " + xy);
				return xy + MAX_STUCK_X;
			}
		}
		return FAR_FROM_STUCK;
	}

    // clicked call means there weren't any drag events.
    public void mouseClicked(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
		int action = getActionMode();
		int stuckValue = closeToStuck();
		if (MOUSEOVER == action && stuckValue != FAR_FROM_STUCK)
		{
//			System.out.println("picked stuck position: " + stuckValue);
			baseBox = createBoxFromCells(stuckValue);
			flashBox = baseBox;
			setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
			setActionMode(MOVE_STUCKPOS);
		}
	else if (RECT_FREEMOVE == getActionMode() || RECT_FREEMOVEAGAIN == getActionMode() ||
	    ST_FREEMOVE == getActionMode() || ST_FREEMOVEAGAIN == getActionMode()) {
	    // throw away return value, we just need the check of
	    // whether faraway click means revert to mouseover
		if (null == baseBox) {
			actionmodeReset();
		}
//		else
//	    	createRectFromPick(dragStart, dragEnd, baseBox);
	      
	}
	mouseNow[X] = e.getX();
	mouseNow[Y] = e.getY();
	int temp = mouseNow[X];
	if (!horiz) {
		int x = getWinMax(X);
		int y = getWinMax(Y);
		mouseNow[X] = mouseNow[Y] * x / y;
		mouseNow[Y] = temp * y / x;
	}
    }

    public void mousePressed(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
	
	mouseDragging = true;
	mouseNow[X] = e.getX();
	mouseNow[Y] = e.getY();
	int temp = mouseNow[X];
	if (!horiz) {
		int x = getWinMax(X);
		int y = getWinMax(Y);
		mouseNow[X] = mouseNow[Y] * x / y;
		mouseNow[Y] = temp * y / x;
	}
	for (int xy = 0; xy < 2; xy++) {
	    dragStart[xy] = mouseNow[xy];
	    dragEnd[xy] = mouseNow[xy];
	    //mousePrev[xy]= dragStart[xy]; // why???
	}
	if (baseBox != null)
		baseBox.setDragPoints(dragStart, dragEnd);
	
	if (MOUSEOVER == getActionMode()) {
		//drawActiveSubtreeBox();	    
		{		
			tj.clearDrawers();
		}
		actionmode = RECT_CREATE;
	}
	if (RECT_FREEMOVEAGAIN == getActionMode()) {
	    setActionMode(RECT_RESHAPE);
	    remakerectangle();
	} else if (RECT_FREEMOVE == getActionMode()) {
	    if (null != flashBox) {
		setActionMode(RECT_RESHAPE);
	    } else {
		// no box yet. drag one out.
		flashBoxWidth=3f;
		setActionMode(RECT_CREATE);
	    }
	} else if (ST_FREEMOVE == getActionMode()) {
	    setActionMode(ST_RESHAPE);
	} else if (ST_FREEMOVEAGAIN == getActionMode()) {
	    setActionMode(ST_RESHAPE);
	    //remakerectangle();
	}
    }

    public void mouseReleased(MouseEvent e) {
		if (RECT_RESHAPE == getActionMode() || ST_RESHAPE == getActionMode())
		{
			baseBox.undraw();

			flashBox.undraw();
			doFlash();
			requestRedraw(); // too much drawing
			actionmodeReset();
//			System.out.println("release");
			return;
		}
//		if (mouseOutOfWindow(e))
//			return;
	mouseDragging = false;
	if (RECT_CREATE == getActionMode()) {
	    setActionMode(RECT_FREEMOVE);
	} else if (RECT_RESHAPE == getActionMode() || MOVE_STUCKPOS == getActionMode()) {
//	    setActionMode(RECT_FREEMOVEAGAIN);
//        setActionMode(MOUSEOVER);
           actionmodeReset();          
//		flashBoxWidth=3f;
//		baseBox = createBoxFromCells();
//		flashBox = baseBox;
//		setFlash(null, tj.rubberbandColor, mouseNow[X], mouseNow[Y], true);		
 //       remakerectangle();
 		
 		
	} else if (getActionMode() == ST_RESHAPE) {
	    setActionMode(ST_FREEMOVEAGAIN);
	}
    }

    public void mouseDragged(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
	mouseNow[X] = e.getX();
	mouseNow[Y] = e.getY();
	int temp = mouseNow[X];
	if (!horiz) {
		int x = getWinMax(X);
		int y = getWinMax(Y);
		mouseNow[X] = mouseNow[Y] * x / y;
		mouseNow[Y] = temp * y / x;
	}
	dragEnd[X] = mouseNow[X];
	dragEnd[Y] = mouseNow[Y];
	if (baseBox != null)
	{
		baseBox.updateDrag(dragEnd);
	} 
	   
	if (mouseNow[X] == mousePrev[X] && 
	    mouseNow[Y] == mousePrev[Y]) return;
		if (ST_FREEMOVE == getActionMode())
	//if (ST_FREEMOVE == actionmode || ST_FREEMOVEAGAIN == actionmode)
	    	setActionMode(ST_RESHAPE);
	if (ST_RESHAPE == getActionMode()) {
	    reshaperectangle(linkednav);
	}
	else if (getActionMode() == MOVE_STUCKPOS)
	{
//		System.out.println("moving stuck position");
		moveStuckPosition(baseBox);
	} else if ((RECT_CREATE == getActionMode())) {
		tj.setQuasimode(true);
		flashBox = null;
		flashBoxWidth=3f;
		int x=getWinMax(0);
		int y=getWinMax(1);
//		if(this.drawDirection == LEFT)
//		{
			baseBox = createBoxFromCells(dragStart, dragEnd);
			flashBox = baseBox;
//		}
//		if(this.drawDirection == RIGHT)
//		{
//			int[] newDragStart = new int[2];
//			int[] newDragEnd = new int[2];
//			newDragStart[0] = this.getWinsize(X)-dragStart[X];
//			newDragStart[1] = dragStart[Y];
//			newDragEnd[0] = this.getWinsize(X)-dragEnd[X];
//			newDragEnd[1] = dragEnd[Y];	
//			baseBox = createBoxFromCells(newDragStart, newDragEnd);
//			flashBox = baseBox;		
//		}	
//		if(this.drawDirection == UP)
//			{
//				int[] newDragStart = new int[2];
//				int[] newDragEnd = new int[2];
//				newDragStart[0] =x-dragStart[Y]*x/y;
//				newDragStart[1] =dragStart[X]*y/x;
//				newDragEnd[0] = x-dragEnd[Y]*x/y;
//				newDragEnd[1] = dragEnd[X]*y/x;		
//				baseBox = createBoxFromCells(newDragStart, newDragEnd);
//				flashBox = baseBox;		
//			}	
//		if(this.drawDirection == DOWN)
//			{
//				int[] newDragStart = new int[2];
//				int[] newDragEnd = new int[2];
//				newDragStart[0] = dragStart[Y]*x/y;
//				newDragStart[1] =y-dragStart[X]*y/x;
//				newDragEnd[0] = dragEnd[Y]*x/y;
//				newDragEnd[1] = y-dragEnd[X]*y/x;		
//				baseBox = createBoxFromCells(newDragStart, newDragEnd);
//				flashBox = baseBox;		
//			}	
	    	    
//               baseBox = createBoxFromCells(dragStart, dragEnd);
//              flashBox = baseBox;
//		     setFlash(	null,rubberbandColor,mouseNow[X],mouseNow[Y],true);			
              setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
		    //  setActionMode(RECT_FREEMOVE);	
	} else if (RECT_RESHAPE == getActionMode()) {
			    reshaperectangle(false);
	}
	
	// modification by jeffrey, state machine change for control key vs. drag start
	
	  else if(RECT_FREEMOVE == getActionMode()){
		  flashBoxWidth=3f;
		  baseBox = createBoxFromCells(dragStart, dragEnd);
		  flashBox = baseBox;
		  setFlash(null, rubberbandColor, e.getX(), e.getY(), true);        
	  }
	mousePrev[X] = mouseNow[X];
	mousePrev[Y] = mouseNow[Y];
    }

    public void mouseMoved(MouseEvent e) {
//		System.out.println("mousemoved");
		if (mouseOutOfWindow(e))
			return;
    	
	mouseNow[X] = e.getX();
	mouseNow[Y] = e.getY();
	int temp = mouseNow[X];
	int x = getWinMax(X);
	int y = getWinMax(Y);
	if (!horiz) {		
		mouseNow[X] = mouseNow[Y] * x / y;
		mouseNow[Y] = temp * y / x;
	}
	if (MOUSEOVER == getActionMode())
	{
//	    if(this.drawDirection==LEFT)
	    mouseover(e.getX(), e.getY());
//	    if(this.drawDirection==RIGHT)
//	    mouseover(this.getWinsize(0)-e.getX(), e.getY());
//	    if(this.drawDirection==UP)
//		mouseover(x-e.getY()*x/y, e.getX()*y/x);
//	    if(this.drawDirection==DOWN)
//	    mouseover(e.getY()*x/y,y-e.getX()*y/x);
	}
	mousePrev[X] = mouseNow[X];
	mousePrev[Y] = mouseNow[Y];
    }

    /** Used when hitting a key should trigger an atomic action.
     * Alternative is a {@link keyPressed() keyPressed}/{@link
     * keyReleased() keyReleased} pair that could be used in
     * conjunction with mouseclicks.
     * @author Tamara Munzner */
    public void keyTyped(KeyEvent e) {
	char key = e.getKeyChar();

	if (e.isShiftDown())
	{
		tj.setQuasimode(false);
	    actionmodeReset();
		System.out.println("upper case?");
	}
	System.out.println("you typed " + key + " kt + " + keytarget);
	if (key >= '0' && key <= '3') {
	    keytarget = TreeJuxtaposer.markGroup[key - '0'];
//	    System.out.println("keytarget " + (key-'0') + " bravo");
	} else if (key >= '4' && key <= '9') {
		System.out.println("Are you sure? " + key + " should be between 0 and 3, inclusive");
	} else if (key == 'f') {
	    keytarget = TreeJuxtaposer.flashGroup;
	} else if (key == 'd') {
	    keytarget = TreeJuxtaposer.diffGroup;
	} else if (key == 'l') {
	    keytarget = TreeJuxtaposer.foundGroup;
	} else if (key == 'h') {
	    keytarget = HORIZ;
	    setGrowDirection(HORIZ);
	} else if (key == 'v') {
	    keytarget = VERT;
	    setGrowDirection(VERT);
	} else if (key == 'a') {
	    keytarget = ALL;
	    setGrowDirection(ALL);
	    setSelectionResolution(SUBTREE);
	} else if (key == 'n') {
	    keytarget = NODE;
	    setSelectionResolution(NODE);
	} else if (key == '.' || key == '>') { // with or w/o shift
	    keytarget = DEVELOPER;
	} else {
	    if (keytarget != DEVELOPER) {
		if (key == 'g') {
		    if (keytarget == TreeJuxtaposer.markGroup[0] ||
		    keytarget == TreeJuxtaposer.markGroup[1] ||
			keytarget == TreeJuxtaposer.markGroup[2] ||
			keytarget == TreeJuxtaposer.markGroup[3])
			{
				setMarkGroup(keytarget);
//				System.out.println("set group to " + (keytarget) + " all by myself");
			}
		    else {
		    	int newMark = getMarkGroup() + 1;
			if (newMark > TreeJuxtaposer.markGroup[3]) 
				setMarkGroup(TreeJuxtaposer.markGroup[0]);
			else
				setMarkGroup(newMark);
				
//			System.out.println("pre Priorities: 0:" + tj.getGroup(TreeJuxtaposer.markGroup[0]).getKey() + " 1:"
//			+ tj.getGroup(TreeJuxtaposer.markGroup[1]).getKey() + " 2:"
//			+ tj.getGroup(TreeJuxtaposer.markGroup[2]).getKey() + " 3:"
//			+ tj.getGroup(TreeJuxtaposer.markGroup[3]).getKey());
		    tj.setGroupPriority(getMarkGroup()); 
//			System.out.println("post Priorities: 0:" + tj.getGroup(TreeJuxtaposer.markGroup[0]).getKey() + " 1:"
//			+ tj.getGroup(TreeJuxtaposer.markGroup[1]).getKey() + " 2:"
//			+ tj.getGroup(TreeJuxtaposer.markGroup[2]).getKey() + " 3:"
//			+ tj.getGroup(TreeJuxtaposer.markGroup[3]).getKey());
			}
			tj.requestRedrawAll();
		} else if (key == 'k') {
		    tj.toggleLinkedNav();
			tj.requestRedrawAll();
		} else if (key == 'r') {
		    tj.reset();
		} else if (key =='R') {
		    reset();
		} else if (key == 't') {
		    if (keytarget == TreeJuxtaposer.markGroup[0] ||
		    	keytarget == TreeJuxtaposer.markGroup[1] ||
		    	keytarget == TreeJuxtaposer.markGroup[2] ||
		    	keytarget == TreeJuxtaposer.markGroup[3] ||
		    	keytarget == TreeJuxtaposer.flashGroup ||
		    	keytarget == TreeJuxtaposer.diffGroup ||
		    	keytarget == TreeJuxtaposer.foundGroup)
		    		setActionGroup(keytarget);
		    else setActionGroup(keytargetDefault);
		    drawActiveSubtreeBox();
		    setActionMode(ST_FREEMOVE);
		    //actionmode = MOUSEOVER;
		    if (getActionGroup() == TreeJuxtaposer.flashGroup) 
			if (pickedNode != null) tj.setFocus(pickedNode, this);
		    else {
			ArrayList forestRoots = tj.getGroupForest(getActionGroup(), this); // array of TreeNodes
			if (!forestRoots.isEmpty()) 
			    tj.setFocus(((TreeNode)forestRoots.get(0)), this);
		    }
		} else if (key == 'b' || key == 'B' || key == 's' || key == 'S') {
			/*
		    if (keytarget == HORIZ || 
			keytarget == ALL) {
			setGrowDirection(keytarget);
		    } else {
			setGrowDirection(VERT);
		    }
		    */
		    boolean grow=(key == 'b' || key == 'B');
		    int numSteps = (jump) ? 1 : getNumAnimSteps();
		    boolean horiz = getGrowDirection() != VERT;
		    boolean vert = getGrowDirection() != HORIZ;
			if (linkednav)
			tj.resizeGroup(tj.getGroup(getActionGroup()), numSteps, grow, horiz, vert);
		    else 
			tj.resizeGroup(tj.getGroup(getActionGroup()), numSteps, grow, horiz, vert, this);
		    if (getActionGroup() == TreeJuxtaposer.flashGroup) 
			if (pickedNode != null) tj.setFocus(pickedNode, this);
		    else {
			ArrayList forestRoots = tj.getGroupForest(getActionGroup(), this); // array of TreeNodes
			if (!forestRoots.isEmpty()) 
			    tj.setFocus(((TreeNode)forestRoots.get(0)), this);
		    }
		    if (linkednav) tj.requestRedrawAll();
		    else requestRedraw();
		} else if (key == 'm' || key == 'M') {
		    if (pickedNode != null) {
		    	/* keytarget was over riding my nice panel implementation - jslack
			if (keytarget == NODE) {
			    setActionTarget(NODE);
			} else {
			    setActionTarget(ALL);
			} 
			*/
			if (key == 'm')	
				// tj.clearGroup(getMarkGroup());  // replace 
		           tj.unmarkGroup(getMarkGroup());
		    else
		    	// silly shift problem
		    	tj.clearDrawers();
			boolean selectSubtree=getSelectionResolution()==SUBTREE;
			tj.doSelectGeom(pickedNode, selectSubtree, getMarkGroup(), this);
			tj.setFocus(pickedNode, this);
			tj.requestRedrawAll();
			tj.lcaNode.add(pickedNode);
		    if(tj.stateFrame.LCAMark.isSelected())
		    	tj.doLCAGeom(TreeJuxtaposer.lcaGroup, this);
//			System.out.println("selected a group: " + tj.getGroup(getMarkGroup()));
			tj.requestRedrawAll();
		    }
		} else if (key == 'c') {
		    if (keytarget == TreeJuxtaposer.markGroup[0] ||
		    keytarget == TreeJuxtaposer.markGroup[1] ||
			keytarget == TreeJuxtaposer.markGroup[2] ||
			keytarget == TreeJuxtaposer.markGroup[3])
			tj.unmarkGroup(keytarget);
			tj.requestRedrawAll();
	    } else if (keytarget == ALL) {
			tj.unmarkGroup(TreeJuxtaposer.markGroup[0]);
			tj.unmarkGroup(TreeJuxtaposer.markGroup[1]);
			tj.unmarkGroup(TreeJuxtaposer.markGroup[2]);
			tj.unmarkGroup(TreeJuxtaposer.markGroup[3]);
			tj.requestRedrawAll();
		    } else {
//			tj.unmarkGroup(getMarkGroup());
		    }
		    tj.clearDrawers();
//		    tj.requestRedrawAll();
	    }
		 else if (key == 'u') {
		    tj.toggleShowDiffs();
		} else if (key == 'w') {
		    tj.toggleDrawLabels();
		} else if (key == 'e') {
		    actionmodeReset();
		} 
	    else {
		System.out.println("second tier");
		 if (key == 'A') {
		    tj.toggleShowAlg();
		} else if (key == 'N') {
		    tj.requestRedrawAll();
		} else if (key == 'b') {
		    tj.toggleBigGrid();
		} else if (key == 'g') {
		    tj.toggleNoGrid();
		} else if (key == 'c') {
		    tj.toggleColorGrid();
		} else if (key == 'T') {
		    tj.toggleNoGeoms();
		    requestRedraw();
		} else if (key == 'H') {
		    tj.toggleNoFlash();
		} else if (key == 'o') {
		    tj.toggleDrawOrder();
		} else if (key == 'k') {
		    tj.toggleExpandLeaves();
		} else if (key == 'R') {
		    tj.toggleLabelPosRight();
		} else if (key == 'w') {
		    tj.increaseLineThickness();
		} else if (key == 'W') {
		    tj.decreaseLineThickness();
		} else if (key == 'm') {
		    tj.increaseLabelBuffer();
		} else if (key == 'M') {
		    tj.decreaseLabelBuffer();
		} else if (key == 'x') {
		    tj.increaseMaxFontHeight();
		} else if (key == 'X') {
		    tj.decreaseMaxFontHeight();
		} else if (key == 'e') {
		    tj.increaseMinFontHeight();		    
		} else if (key == 'E') {
		    tj.decreaseMinFontHeight();		    
		} else if (key == 'B') {
		    tj.toggleLabelDrawBack();
		} else if (key == 'p') {
		    tj.toggleLabelPopup();
		} else if (key == 'D') {
		    tj.toggleDimBrite();
		} else if (key == 'C') {
		    tj.toggleDimColors();
		} else if (key == 'j') {
		    jump = !jump;
		} else if (key == 's') {
		    tj.toggleLabelTransp();
		} else if (key == 'z') {
		    tj.toggleDrawSplits();
		} else if (key == 'O') {
		    tj.toggleCacheRange();
		} else if (key == 'G') {
		    tj.toggleGuarVis();
		} else if (key == 't') {
		    tj.toggleCheckTime();
		} else if (key == 'J') {
		    stats.flush();
		    System.out.println("flushed stats\n");
		} else if (key == 'S') {
		    tj.toggleDumpStats();
		} else if (key == 'F') {
		    // the second display loop is so that the entire scene gets drawn, before going on to next iteration
		    int i;
		    setDumpStats(true);
		    setCheckTime(true);
		    setCacheRange(true);
		    setDrawOrder(BRUTEFORCE);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("brute checktime cache "+i+" ");display();}
		    setCheckTime(false);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("brute nochecktime cache "+i+" ");display();}
		    setDrawOrder(FROMFOCUSCELL);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("focus nochecktime cache "+i+" ");display();}
		    setCheckTime(true);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("focus checktime cache "+i+" "); display(); while (keepDrawing) display();}

		    stats.flush();
		    stats.println();
		    stats.println("MARK WHOLE TREE");
		    stats.println();
		    
		    tj.clearGroup(TreeJuxtaposer.markGroup[0]);
		    tj.doSelectGeom(rootNode, true, TreeJuxtaposer.markGroup[0], this);
		    tj.setFocus(rootNode, this);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("mark checktime cache "+i+" ");display(); while (keepDrawing) display();}
		    setCheckTime(false);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("mark nochecktime cache "+i+" ");display();}
		    setCacheRange(false);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("mark nochecktime nocache "+i+" ");display();}
		    setCheckTime(true);
		    for (i = 0; i < 10; i++) {stats.println();stats.println();stats.print("mark checktime nocache "+i+" ");display(); while (keepDrawing) display();}
		    stats.flush(); stats.println();
		    stats.println();stats.println("DONE WITH STATS");
		    stats.flush();
		    setDumpStats(false);
		    System.out.println("done gathering stats");
		}
	    }
	    tj.requestRedrawAll();
	    keytarget = NONE;
	}
	tj.observe();
    }

	private void newMouseOver(TreeNode originalNode, TreeNode newNode)
	{
		if( newNode == null ) return;
		tj.clearGroup(TreeJuxtaposer.flashGroup);
		//System.out.println("Who is this node: " + originalNode);
		//System.out.println("Who is this node (current tree): " + newNode);
//		String nodeName = newNode.getName();
		tj.addNodesToGroup(newNode.key, newNode.key, TreeJuxtaposer.flashGroup, this);
		tj.doFlashGeom(newNode, TreeJuxtaposer.flashGroup, this, getX(), getY(), true);
		this.changedMarks();
		this.pickedNode = newNode;
	}

	// true if treeNode cells (level 0 of quadtree) overlap horizontally
	private boolean inHorizRange(TreeNode a, TreeNode b)
	{
		final int A = 0, B = 1;
		double min[], max[];
		min = new double[2];
		max = new double[2];
		if (a.parent != null)
			min[A] = a.parent.getCell().getMax(X);
		else
			min[A] = -1f;
		if (b.parent != null)
			min[B] = b.parent.getCell().getMax(X);
		else
			min[B] = -1f;
		max[A] = a.getCell().getMax(X);
		max[B] = b.getCell().getMax(X);
		if (max[A] <= min[B] || max[B] <= min[A])
			return false;
		return true;
	}
	
	// find the child number of the originalNode as a child of the parent node
	private int findChildNumber(TreeNode parentNode, TreeNode originalNode)
	{
		int originalNodeChildNumber = 0; // number of original node (plus 1 actually)
		while (originalNodeChildNumber < parentNode.numberChildren() &&
					parentNode.getChild(originalNodeChildNumber++) != originalNode);
		if (originalNodeChildNumber > parentNode.numberChildren())
		{
			System.out.println("Child couldn't find self from parent");
		}
		return originalNodeChildNumber;
	}
	
	// find the right aligned cell that is below (above) the original node, using the parent node as reference
	// - the last parameter (boolean) is used to determine whether to search down (true) or up (false)
	private TreeNode rightAlignFind(TreeNode parentNode, TreeNode originalNode, int originalNodeChildNumber, boolean goingDown)
	{
		TreeNode temp = parentNode.getChild(originalNodeChildNumber);
		TreeNode prev = temp;
		final int TEMP = 0;
		final int ORIG = 1;
		double rightX[];
		rightX = new double[2];
		rightX[ORIG] = originalNode.getMaxY();
		rightX[TEMP] = temp.getMaxY();
//		System.out.println("Firstly: ");
//		System.out.println("rightX[ORIG] = " + rightX[ORIG]);
//		System.out.println("rightX[TEMP] = " + rightX[TEMP]);			  
		while (rightX[TEMP]  < rightX[ORIG] && temp.numberChildren() > 0)
		{
			prev = temp;
			if (goingDown)
				temp = temp.getChild(0);
			else
				temp = temp.getChild(temp.numberChildren() - 1);
			rightX[TEMP] = temp.getMaxY();
//			System.out.println("rightX[TEMP] = " + rightX[TEMP]);
		}
		if (rightX[TEMP] == rightX[ORIG] || temp.numberChildren() == 0)
			return temp;
		else
			return prev;
	//return null;
	}
	//	mouseover parent
	// do nothing if currently no mouseover or at root node
	public void flashLeft()
	{
		if (tj.getGroup(TreeJuxtaposer.flashGroup).size() == 0)
			return;
		RangeInTree rangeInTree =
			(RangeInTree) tj.getGroup(TreeJuxtaposer.flashGroup).getFirst();
		int firstNodeKey = rangeInTree.getMin();
		TreeNode originalNode, parentNode;
		originalNode = tree.getNodeByKey(firstNodeKey);
		if (originalNode == null
			|| (parentNode = originalNode.parent()) == null)
			return;
		mouseOverNode = parentNode; // keep the level for up/down traversal
		newMouseOver(originalNode, parentNode);
	}
	
	//	mouseover first child 
	// do nothing if currently no mouseover or at terminal node
	public void flashRight()
	{
		if (tj.getGroup(TreeJuxtaposer.flashGroup).size() == 0)
			return;
		RangeInTree rangeInTree =
			(RangeInTree) tj.getGroup(TreeJuxtaposer.flashGroup).getFirst();
		int firstNodeKey = rangeInTree.getMin();
		TreeNode originalNode, childNode = null;
		originalNode = tree.getNodeByKey(firstNodeKey);
		if (originalNode == null
			|| originalNode.numberChildren() == 0
			|| (childNode = originalNode.getChild(0)) == null)
			return;
		mouseOverNode = childNode; // keep the level for up/down traversal
		newMouseOver(originalNode, childNode);
	}


	//	mouseover sibling if at first child, cousin if at last child of common parent
	// do nothing if at bottommost node (bottommost has no tree edges below that share X coord)
	public void flashDown()
	{
//		RangeInTree rangeInTree =
//			(RangeInTree) tj.getGroup(tj.flashGroup).ranges.getFirst();
		RangeInTree rangeInTree =
			(RangeInTree) tj.getGroup(TreeJuxtaposer.flashGroup).getFirst();
		int firstNodeKey = rangeInTree.getMin();
		TreeNode prevParentNode = null,
			parentNode = null,
			originalNode,
			downNode = null;
		originalNode = tree.getNodeByKey(firstNodeKey);
		downNode = originalNode; // default downNode, idempotent action
		if (originalNode == null
			|| (parentNode = originalNode.parent()) == null)
			return; // no parent (root) or no highlight
		int originalNodeChildNumber = findChildNumber(parentNode, originalNode);
		// number of original node (plus 1 actually)
		if (originalNodeChildNumber > parentNode.numberChildren())
			return; // can't find self
		if (originalNodeChildNumber < parentNode.numberChildren()) {
			// become maximal bigger-numbered-sibling's descendent
			//  - of the immediate parent, find the next sibling and traverse down
			//    until the last node that has a right edge less than the right edge of
			//    the original node is found (so traversal through the leaves of a right-
			//    aligned tree is possible since it is better than falling back to the root)
			downNode =
				rightAlignFind(
					parentNode,
					mouseOverNode,
					originalNodeChildNumber,
					true);
		} else {
			// process cousins
			prevParentNode = parentNode.parent;
			int parentAsChildNum = 0;
			while (prevParentNode != null) {
				parentAsChildNum = findChildNumber(prevParentNode, parentNode);
				if (parentAsChildNum > prevParentNode.numberChildren())
					return;
				if (parentAsChildNum < prevParentNode.numberChildren()
					|| prevParentNode.parent == null)
					break;
				parentNode = prevParentNode;
				prevParentNode = prevParentNode.parent;
			}
			if (prevParentNode == null
				|| parentAsChildNum == prevParentNode.numberChildren())
				return;
			downNode =
				rightAlignFind(
					prevParentNode,
					mouseOverNode,
					parentAsChildNum,
					true);
		}
		newMouseOver(originalNode, downNode);
	}

	//	mouseover sibling if at last child, cousin if at first child of common parent
	// do nothing if at topmost node
	public void flashUp()
	{
		RangeInTree rangeInTree = tj.getGroup(TreeJuxtaposer.flashGroup).getFirst();
		int firstNodeKey = rangeInTree.getMin();
		TreeNode prevParentNode = null,
			parentNode = null,
			originalNode,
			upNode = null;
		originalNode = tree.getNodeByKey(firstNodeKey);
		upNode = originalNode; // default upNode, idempotent action
		if (originalNode == null || (parentNode = originalNode.parent()) == null)
			return; // no parent (root) or no highlight
		int originalNodeChildNumber = findChildNumber(parentNode, originalNode);
		// number of original node (plus 1 actually)
		if (originalNodeChildNumber > parentNode.numberChildren())
			return; // can't find self
		if (originalNodeChildNumber > 1) {
			// become maximal smaller-numbered-sibling's descendent
			//  - of the immediate parent, find the previous sibling and traverse up
			//    until the last node that has a right edge less than the right edge of
			//    the original node is found (so traversal through the leaves of a right-
			//    aligned tree is possible since it is better than falling back to the root)
			upNode =
				rightAlignFind(
					parentNode,
					mouseOverNode,
					originalNodeChildNumber - 2,
					false);
		} else {
			// process cousins
			prevParentNode = parentNode.parent;
			int parentAsChildNum = 0;
			while (prevParentNode != null) {
	
				parentAsChildNum = findChildNumber(prevParentNode, parentNode);
				if (parentAsChildNum > prevParentNode.numberChildren())
					return; // can't find parent from prev parent
				if ((parentAsChildNum > 1
					&& parentAsChildNum <= prevParentNode.numberChildren())
					|| prevParentNode.parent == null)
					break; // found good common descendent to traverse down
				parentNode = prevParentNode;
				prevParentNode = prevParentNode.parent;
			}
			if (prevParentNode == null || parentAsChildNum == 1)
				return;
			upNode =
				rightAlignFind(
					prevParentNode,
					mouseOverNode,
					parentAsChildNum - 2,
					false);
			//0, false);
		}
		newMouseOver(originalNode, upNode);
	}

    public void keyPressed(KeyEvent e) {
//	if (keytarget == DEVELOPER) return;
	if (tj.getQuasimode()) return;
	char key = e.getKeyChar();
	int code = e.getKeyCode();
//	if (keytarget == DEVELOPER)
//	{
//	    if (code == KeyEvent.VK_LEFT)			     			    
//			{
//					tj.drawLL(); 
//					tj.butterflyFrame.setButtonStatus(0);
//			} 		
//		 else if (code == KeyEvent.VK_RIGHT)   
//			{
//				tj.drawRR();
//				tj.butterflyFrame.setButtonStatus(1);
//			}  
//		else if (code == KeyEvent.VK_DOWN)
//			{
//				// don't support down right now
//					//tj.drawDD();
//				//tj.butterflyFrame.setButtonStatus(4);
//			}		
//		else if (code == KeyEvent.VK_UP)
//			{
//					tj.drawUU(); 
//				    tj.butterflyFrame.setButtonStatus(3);
//			} 
//		else if(key == '2')
//		{
//		        tj.drawRL();
//			tj.butterflyFrame.setButtonStatus(2);
//		}
//	}
 //   else {
	if (code == KeyEvent.VK_LEFT)
	{
//	   if(this.drawDirection == LEFT)
		flashLeft();
//		if(drawDirection == RIGHT)
//		flashRight();
//		if(drawDirection == UP)
//		flashUp();
//		if(drawDirection == DOWN)
//		flashDown();
	}
	else if (code == KeyEvent.VK_RIGHT)
	{
//		if(drawDirection == LEFT)
		flashRight();
//		if(drawDirection == RIGHT)
//		flashLeft();
//		if(drawDirection == UP)
//		flashDown();
//		if(drawDirection == DOWN)
//		flashUp();
	}
	else if (code == KeyEvent.VK_DOWN)
	{
//		   if(drawDirection == LEFT || drawDirection == RIGHT)
			flashDown();
//			if(drawDirection == UP)
//			flashLeft();
//			if(drawDirection == DOWN)
//			flashRight();
	}
	
	else if (code == KeyEvent.VK_UP)
		{
//			if(drawDirection == LEFT || drawDirection == RIGHT)
			 flashUp();
//			 if(drawDirection == UP)
//			 flashRight();
//			 if(drawDirection == DOWN)
//			 flashLeft();
		}
	else 
	//System.out.println("PRESS " + key + " code " + code);
	if (KeyEvent.VK_CONTROL == code || KeyEvent.VK_ALT == code) {
	    tj.setQuasimode(true);
	    flashBox = null;
	    setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
	    setActionMode(RECT_FREEMOVE);
	}  else if (KeyEvent.VK_SHIFT == code) {
	    if (keytarget == TreeJuxtaposer.markGroup[0] ||
	    	keytarget == TreeJuxtaposer.markGroup[1] ||
	    	keytarget == TreeJuxtaposer.markGroup[2] ||
	    	keytarget == TreeJuxtaposer.markGroup[3] ||
	    	keytarget == TreeJuxtaposer.flashGroup ||
	    	keytarget == TreeJuxtaposer.diffGroup ||
	    	keytarget == TreeJuxtaposer.foundGroup)
	    		setActionGroup(keytarget);
	    else setActionGroup(keytargetDefault);
	    drawActiveSubtreeBox();
	    if (getActionGroup() == TreeJuxtaposer.flashGroup) {
		if (null != pickedNode) {
		    tj.setFocus(pickedNode, this);
		    tj.setQuasimode(true);
		    drawActiveSubtreeBox();
		    setActionMode(ST_FREEMOVE);
		}
	    } else {
		tj.setQuasimode(true);
		drawActiveSubtreeBox();
		setActionMode(ST_FREEMOVE);
		ArrayList forestRoots = tj.getGroupForest(getActionGroup(), this); // array of TreeNodes
		if (!forestRoots.isEmpty()) 
		    tj.setFocus(((TreeNode)forestRoots.get(0)), this);
	    }
	    keytarget = NONE;
	}
  //  }
    }

    public void keyReleased(KeyEvent e) {
	if (keytarget == DEVELOPER) return;
	char key = e.getKeyChar();
	int code = e.getKeyCode();
	if (KeyEvent.VK_SHIFT == code || KeyEvent.VK_CONTROL == code || KeyEvent.VK_ALT == code) {
	    tj.setQuasimode(false);
	    actionmodeReset();
	}
	//System.out.println("RELEASE " + key + " code " + code);
    }


public void seedQueue() {
//	System.out.println("Seeding queue for: " + this);
    tj.seedQueue(this);
//    System.out.println("Seeded: " + ToDrawQ);
}

	public ArrayList getColorsForCellGeom(CellGeom c){
		return tj.getColorsForRange(c.getKey(), c.getKey(), this);
	}


	/**
	 * @return
	 */
	public InteractionBox getBaseBox() {
		return baseBox;
	}
	/* (non-Javadoc)
	 * @see AccordionTreeDrawer.AccordionTreeDrawer#getRangeByKey(int)
	 */
	public RangeList getRangeByKey(int key)
	{
		if (key >= 0 && key < tj.groups.size())
			return tj.getGroup(key);
		return null;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#getColorsForGridCell(AccordionDrawer.GridCell, int)
	 */
	public ArrayList getColorsForGridCell(GridCell c, int objkey)
	{
		return null;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#getOlduvaiObject()
	 */
	public OlduvaiObject getOlduvaiObject()
	{
		return (Tree)tree;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#drawReferenceLine()
	 */
	public void drawReferenceLine()
	{	
	}

};
